/*
 * Copyright (c) 2002 Regents of The University of Michigan.
 * All Rights Reserved.  See COPYRIGHT.
 */



#define _WIN32_WINNT 0x0500

#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include "sparse.h"

#include <openssl\ssl.h>
#include <openssl\err.h>
#include "libsnet/snet.h"
#include "CosignFilterMain.h"
#include "CosignStructs.h"
#include "config.h"


int
CheckCookie( char *, struct sinfo *, struct connlist **, SSL_CTX*, int, char *, int, SERVICECONFIGDATA* );

int
CookieValid( char* szCookie, struct connlist **c_cur, SERVICECONFIGDATA* scd, COSIGNSETTINGS* cs, struct sinfo *si, char* szIpAddr, SSL_CTX* ctx ) {

	struct sinfo	lsi;
	char	szFileName[ MAX_PATH ];
	char	szTempPath[ MAX_PATH ];
	char	szBuffer[ DEFAULTBUFFERSIZE ];
	int		rs;
	int		rc;
	int    writetkt = FALSE;
	FILETIME	curTime;
	ULARGE_INTEGER	ulTime;
	HANDLE		hCookieFile;
	HANDLE		hTempFile;
	DWORD		dwBytesWritten;

	memset((void *)&lsi, 0, sizeof(lsi)); // zero out contents of lsi

	if ( strchr( szCookie, '/' ) != NULL ) {
		CosignLog( CS_ERROR, "Cookie contains '/'" );
		return( COSIGN_ERROR );
	}
	//check for buffer overrun here of szFileName
	//The '+2' is for the back slash and terminating NULL character.
    if ( strlen( cs->cs_cookieDBPath ) + strlen( szCookie ) + 2 > MAX_PATH ) {
		CosignLog( CS_ERROR, "Buffer for cookie path is too small.\n" );
		return( COSIGN_ERROR);
	}
	if ( _snprintf( szFileName, sizeof( szFileName ), "%s\\%s", cs->cs_cookieDBPath, szCookie ) > sizeof( szFileName ) ) {
		///Should probably make buffer bigger?
        CosignLog( CS_ERROR, "Buffer for cookie path + cookie name is too small.\n" );
	}
	GetSystemTimeAsFileTime( &curTime );
	CSDEBUG1( CS_INFO, "ReadSCookie( %s )...\n",  szFileName );
	if ( ( rs = ReadSCookie( szFileName, &lsi ) ) < 0 ) {
		CosignLog( CS_ERROR, "ReadSecant() failed." );
		return( COSIGN_ERROR );
	}
	ConvertFTtoUl( &curTime, &ulTime );
	if ( ( rs == 0 )
        && ( ulTime.QuadPart - lsi.si_itime.QuadPart <= cs->cs_cookieDBExpireTime )
	) {

		strcpy( si->si_ipaddr, lsi.si_ipaddr );
		strcpy( si->si_user, lsi.si_user );
		strcpy( si->si_realm, lsi.si_realm );
        strcpy( si->si_factor, lsi.si_factor );

        if ( cs->cs_getkrbtkt ) {
			strcpy( si->si_krb5tktfile, lsi.si_krb5tktfile );
		}
		//Only check the IP address if it is required.
		//If szIpAddr is NULL (length <= 0) then the check IP address option
		//is turned off.
        if ( strcmp( szIpAddr, lsi.si_ipaddr ) != 0 ) {
            CosignLog( CS_ERROR, "Cached IP %s does not match browser IP %s.\n", szIpAddr, lsi.si_ipaddr  );
			if ( cs->cs_checkIpAddr == always ) {
				return( COSIGN_ERROR );
			}
			// Store new ip address value!
			goto storecookie;
        }
		
		return ( COSIGN_OK );
	}

	if ( rs == 1 && cs->cs_getkrbtkt ) {
        writetkt = TRUE;
    }

	if ( _snprintf( szBuffer, sizeof( szBuffer ), "%s=%s", scd->scd_name, szCookie ) > sizeof( szBuffer ) ) {
		///Automatically make buffer bigger?
        CosignLog( CS_ERROR, "Buffer cookie is too small.\n" );
		return( COSIGN_ERROR );
	}
	if ( ctx == NULL ) {
	    CSDEBUG0( CS_INFO, "CookieValid, ctx is NULL.\n" );
	}
    if ( ( rc = CheckCookie( szBuffer, si, c_cur, ctx, cs->cs_getkrbtkt, cs->cs_krbTicketDir, cs->cs_protocolVersionNeeded, scd ) ) == COSIGN_ERROR ) {
		CosignLog( CS_ERROR, "CookieValid::CheckCookie() failed\n" );
		return( COSIGN_ERROR );
	}
	if ( rc == COSIGN_RETRY ) {
		return( COSIGN_RETRY );
	}
    if ( rc == COSIGN_LOGGED_OUT ) {
        CSDEBUG0( CS_INFO, "CookieValid is returning COSIGN_LOGGED_OUT\n" );
        return( COSIGN_LOGGED_OUT );
    }
    if ( rc == COSIGN_BAD_SUFFIX ) {
        return( COSIGN_BAD_SUFFIX );
    }


	if ( strcmp( szIpAddr, si->si_ipaddr ) != 0 ) {
		CosignLog( CS_ERROR, "Server IP info for client %s does not match browser IP %s.\n", szIpAddr, lsi.si_ipaddr  );
		if ( cs->cs_checkIpAddr == always ) {
			CSDEBUG0( CS_INFO, "Failed always ip check, denying access.\n" );
			return( COSIGN_ERROR );
		}
	}

/*
There are three ip address, the szIpAddr, which is the client's current ip.
the lsi.si_ipaddr, which is the cached cookie's ip address of the client.
si->si_ipaddr, which is the ip address the client had when signing in.
*/



	if ( rs == 0 ) {
		if ( strcmp( si->si_user, lsi.si_user ) != 0 ) {
			CosignLog( CS_ERROR, "Network info for username %s does not match local info %s for cookie %s\n",
				si->si_user, lsi.si_user, szCookie );
			return( COSIGN_ERROR );
		}
		if ( strcmp( si->si_realm, lsi.si_realm ) != 0 ) {
			CosignLog( CS_ERROR, "Network info for realm %s does not match local info %s for cookie %s\n",
				si->si_user, lsi.si_user, szCookie );
			return( COSIGN_ERROR );
		}
		if ( ( strcmp( si->si_ipaddr, lsi.si_ipaddr ) != 0 ) ) {
			CosignLog( CS_ERROR, "Network info %s does not match local info %s for cookie %s\n",
                 si->si_ipaddr, lsi.si_ipaddr, szCookie );
			goto storecookie;
			// We don't care if the locally cached cookie does not match the network info.  What's important is if
			// the network info (si->si_ipaddr) does or does not match the browser's ip address (szIpAddr), which is
			// checked later, before the cookie is cached in the file system (cookiedb).
		}
		if ( ( strcmp( si->si_factor, lsi.si_factor ) != 0 ) ) {
			goto storecookie;
		}
        if ( cs->cs_getkrbtkt ) {
		    strcpy( si->si_krb5tktfile, lsi.si_krb5tktfile );
		}

		// Since the network and local info match, update the cached cookie's timestamp.
		GetSystemTimeAsFileTime( &curTime );
		hCookieFile = CreateFile( szFileName, GENERIC_WRITE, FILE_SHARE_READ, NULL,
			OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
		if ( hCookieFile == INVALID_HANDLE_VALUE ) {
			CosignLog( GetLastError(), "CookieValid::CreateFile( %s ) returned INVALID_HANDLE_VALUE.  Could not update cookiedb.\n", szFileName );
			return( COSIGN_ERROR);
		}
		SetFileTime( hCookieFile, NULL, NULL, &curTime );
		if ( hCookieFile != INVALID_HANDLE_VALUE ) {
			CloseHandle( hCookieFile );
		}
		return( COSIGN_OK );
	}

storecookie:
    
    if ( rs == 1 && strcmp( szIpAddr, si->si_ipaddr ) != 0 ) {
		CSDEBUG0( CS_INFO, "Doing initial ip check.\n" );
		CosignLog( CS_INFO, "Client's IP changed from %s to %s\n", szIpAddr, si->si_ipaddr );
		// if rs == 1, it means this cookied hasn't been cached and we should do an initial IP check, if necessary.
		if ( ( cs->cs_checkIpAddr == initial) || (cs->cs_checkIpAddr == always) ) {
			return( COSIGN_ERROR );
		}
		// If not doing IP check, then just log the change and carry on.
	}

	/* store local copy of secant */
	/* First create a temp file so if anything goes wrong we don't corrupt
	 * the current cookie file. */
	//cck for "Cosign CooKie"
    if ( GetTempFileName( cs->cs_cookieDBPath, "cck", 0, szTempPath ) == 0 ) {
		CosignLog( GetLastError(), "CookieValid::GetTempFileName( %s ) failed.\n", cs->cs_cookieDBPath );
		return( COSIGN_ERROR);
	}
	/* GetTempFileName guarantees a unique file name so CREATE_ALWAYS flag
	 * should be ok (i.e., not overwrite another temporary cookie file) */
	if ( ( hTempFile = CreateFile( szTempPath, GENERIC_WRITE, FILE_SHARE_READ, NULL,
		CREATE_ALWAYS, FILE_ATTRIBUTE_TEMPORARY, NULL ) ) == INVALID_HANDLE_VALUE ) {
		CosignLog( GetLastError(), "CookieValid::CreateFile( %s ) failed.\n", szTempPath );
		return( COSIGN_ERROR);
	}
	if (  _snprintf( szBuffer, sizeof( szBuffer ), "i%s\n", si->si_ipaddr ) > sizeof( szBuffer ) ) {
		///Buffer overrun, shouldn't happen, but if it does...
		return( COSIGN_ERROR);
	}
	if ( WriteFile( hTempFile, szBuffer, (DWORD)strlen( szBuffer ), &dwBytesWritten, NULL ) == 0 ) {
		if ( hTempFile != INVALID_HANDLE_VALUE ) CloseHandle( hTempFile );
		DeleteFile( szTempPath );
		CosignLog( GetLastError(), "CookieValid::WriteFile(i, %s ) failed.\n" );
		return( COSIGN_ERROR);
	}
	if ( _snprintf( szBuffer, sizeof( szBuffer ), "p%s\n", si->si_user ) > sizeof( szBuffer ) ) {
		CosignLog( CS_ERROR, "Buffer for username is too small.\n" );
		return( COSIGN_ERROR);
	}
	if ( WriteFile( hTempFile, szBuffer, (DWORD)strlen( szBuffer ), &dwBytesWritten, NULL ) == 0 ) {
		if ( hTempFile != INVALID_HANDLE_VALUE ) CloseHandle( hTempFile );
		DeleteFile( szTempPath );
		CosignLog( GetLastError(), "CookieValid::WriteFile(p(rincipal)) failed.\n");
		return( COSIGN_ERROR);
	}
	if ( _snprintf( szBuffer, sizeof( szBuffer ), "r%s\n", si->si_realm ) > sizeof( szBuffer ) ) {
		CosignLog( CS_ERROR, "Buffer for realm is too small.\n" );
		return( COSIGN_ERROR);
	}
	if ( WriteFile( hTempFile, szBuffer, (DWORD)strlen( szBuffer ), &dwBytesWritten, NULL ) == 0 ) {
		if ( hTempFile != INVALID_HANDLE_VALUE ) CloseHandle( hTempFile );
		DeleteFile( szTempPath );
		CosignLog( GetLastError(), "CookieValid::WriteFile(r(ealm)) failed.\n");
		return( COSIGN_ERROR );
    }
    if ( _snprintf( szBuffer, sizeof( szBuffer ), "f%s\n", si->si_factor ) > sizeof( szBuffer ) ) {
		CosignLog( CS_ERROR, "Buffer for factors is too small.\n" );
		return( COSIGN_ERROR);
	}
    if ( WriteFile( hTempFile, szBuffer, (DWORD)strlen( szBuffer ), &dwBytesWritten, NULL ) == 0 ) {
		if ( hTempFile != INVALID_HANDLE_VALUE ) CloseHandle( hTempFile );
		DeleteFile( szTempPath );
		CosignLog( GetLastError(), "CookieValid::WriteFile(f(actors)) failed.\n");
		return( COSIGN_ERROR );
    }
    if ( writetkt ) {
		if ( _snprintf( szBuffer, sizeof( szBuffer ), "k%s\n", si->si_krb5tktfile ) > sizeof( szBuffer ) ) {
			CosignLog( CS_ERROR, "Buffer for kerberos ticket file path is too small.\n" );
			return( COSIGN_ERROR);
		}
		if ( WriteFile( hTempFile, szBuffer, (DWORD)strlen( szBuffer ), &dwBytesWritten, NULL ) == 0 ) {
			CloseHandle( hTempFile );
			DeleteFile( szTempPath );
			CosignLog( GetLastError(), "CookieValid::WriteFile(k(5ticket)) failed.\n");
			return( COSIGN_ERROR);
		}
	}

	if ( hTempFile != INVALID_HANDLE_VALUE ) CloseHandle( hTempFile );
	if ( CopyFileEx( szTempPath, szFileName, NULL, NULL, FALSE, 0 ) == 0 ) {
		CosignLog( GetLastError(), "CookieValid::CopyFileEx( %s, %s ) failed.\n", szTempPath, szFileName );
		return( COSIGN_ERROR);
	}
	DeleteFile( szTempPath );

	return( COSIGN_OK );
}
