/*
 * Copyright (c) 1998 Regents of The University of Michigan.
 * All Rights Reserved.  See COPYRIGHT.
 */

//#include <sys/stat.h>
//#include <sys/param.h>
//#include <sys/types.h>
//#include <syslog.h>
//#include <stdio.h>
//#include <stdlib.h>


#include "sparse.h"
#include "log.h"

#define MAXLEN 256
#define READBUFFERSIZE 256

/*
 * -1 system error
 *  0 OK
 *  1 Service cookie file not in file system
 */
	int
ReadSCookie( char* szPath, struct sinfo *si )
{
	HANDLE	hCookieFile;
	char	lpBuffer[ READBUFFERSIZE ];
	char*	p;
	DWORD	dwBytesRead;
	FILETIME	ftTime;
	
	hCookieFile = CreateFile( szPath, GENERIC_READ, FILE_SHARE_READ, NULL,
		OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

	if ( hCookieFile == INVALID_HANDLE_VALUE ) {
		CosignLog( GetLastError(), "ReadSecant::CreateFile( %s ) failed.\n", szPath );
		return( 1 );
	}
	if ( GetFileTime( hCookieFile, NULL, NULL, &ftTime ) == 0 ) {
		CosignLog( GetLastError(), "ReadSecant::GetFileTime( %s ) failed.\n", szPath );
		if ( hCookieFile != INVALID_HANDLE_VALUE ) CloseHandle( hCookieFile );
		return( -1 );
	}
	ConvertFTtoUl( &ftTime, &si->si_itime );
	while( ReadFile( hCookieFile, lpBuffer, READBUFFERSIZE,
		&dwBytesRead, NULL ) != 0 ) {

		if ( dwBytesRead == 0 ) {
			//Reached EOF
			break;
		}
		if ( lpBuffer[ dwBytesRead - 1 ] != '\n' ) {
			if ( hCookieFile != INVALID_HANDLE_VALUE ) CloseHandle( hCookieFile );
			CosignLog( CS_ERROR, "ReadSecant(): %s, buffer not large enough to read data.", szPath );
			return( -1 );
		}
		lpBuffer[ dwBytesRead - 1 ] = '\0';
		
		p = strtok( lpBuffer, "\n" );
		while ( p != NULL ) {
			switch( p[ 0 ] ) {

			case 'i':
				strcpy( si->si_ipaddr, ++p );
				break;
			
			case 'p':
				strcpy( si->si_user, ++p );
				break;

			case 'r':
				strcpy( si->si_realm, ++p );
				break;

			case 'k':
				strcpy( si->si_krb5tktfile, ++p );
				break;

            case 'f':
                strcpy( si->si_factor, ++p );
                break;

			default:
				OutputDebugString( "ReadSecant(): Unknown keyword" );
				if ( hCookieFile != INVALID_HANDLE_VALUE ) CloseHandle( hCookieFile );
				return( -1 );
			}
			p = ( strtok( NULL, "\n" ) ) ;
		}
	}

	if ( hCookieFile == INVALID_HANDLE_VALUE || CloseHandle( hCookieFile ) == 0 ) {
		CosignLog( GetLastError(), "ReadSecant::CloseHandle( %s ) failed.\n", szPath );
	}

	if ( si->si_ipaddr == NULL ||
		si->si_user == NULL ||
		si->si_realm == NULL ) {
			return( -1 );
	}
	return( 0 );
}

	void
ConvertFTtoUl( FILETIME*	ftTime, ULARGE_INTEGER*	uli )
{

	uli->HighPart = ftTime->dwHighDateTime;
	uli->LowPart = ftTime->dwLowDateTime;
}

